/*
 * Souffle - A Datalog Compiler
 * Copyright (c) 2013, 2014, Oracle and/or its affiliates. All rights reserved
 * Licensed under the Universal Permissive License v 1.0 as shown at:
 * - https://opensource.org/licenses/UPL
 * - <souffle root>/licenses/SOUFFLE-UPL.txt
 */

/************************************************************************
 *
 * @file AbstractExistenceCheck.h
 *
 * Defines a class for evaluating conditions in the Relational Algebra
 * Machine.
 *
 ***********************************************************************/

#pragma once

#include "ram/Condition.h"
#include "ram/Expression.h"
#include "ram/Node.h"
#include "ram/NodeMapper.h"
#include "ram/Relation.h"
#include "souffle/utility/ContainerUtil.h"
#include "souffle/utility/StreamUtil.h"
#include <cassert>
#include <memory>
#include <sstream>
#include <string>
#include <utility>
#include <vector>

namespace souffle::ram {

/**
 * @class AbstractExistenceCheck
 * @brief Abstract existence check for a tuple in a relation
 */
class AbstractExistenceCheck : public Condition {
public:
    AbstractExistenceCheck(Own<RelationReference> relRef, VecOwn<Expression> vals)
            : relationRef(std::move(relRef)), values(std::move(vals)) {
        assert(relationRef != nullptr && "Relation reference is a nullptr");
        for (const auto& v : values) {
            assert(v != nullptr && "value is a nullptr");
        }
    }

    /** @brief Get relation */
    const Relation& getRelation() const {
        return *relationRef->get();
    }

    /**
     *  @brief Get arguments of the tuple/pattern
     *  A null pointer element in the vector denotes an unspecified
     *  pattern for a tuple element.
     */
    const std::vector<Expression*> getValues() const {
        return toPtrVector(values);
    }

    std::vector<const Node*> getChildNodes() const override {
        std::vector<const Node*> res = {relationRef.get()};
        for (const auto& cur : values) {
            res.push_back(cur.get());
        }
        return res;
    }

    void apply(const NodeMapper& map) override {
        relationRef = map(std::move(relationRef));
        for (auto& val : values) {
            val = map(std::move(val));
        }
    }

protected:
    void print(std::ostream& os) const override {
        os << "("
           << join(values, ",",
                      [](std::ostream& out, const Own<Expression>& value) {
                          if (!value) {
                              out << "_";
                          } else {
                              out << *value;
                          }
                      })
           << ") ∈ " << getRelation().getName();
    }

    bool equal(const Node& node) const override {
        const auto& other = static_cast<const AbstractExistenceCheck&>(node);
        return equal_ptr(relationRef, other.relationRef) && equal_targets(values, other.values);
    }

    /** Relation */
    Own<RelationReference> relationRef;

    /** Pattern -- nullptr if undefined */
    VecOwn<Expression> values;
};

}  // namespace souffle::ram
